import torch
from math import sqrt
from mcmc_samplers import Proposal, Sample
from typing import Tuple, Union

"""
Defines the `GaussianRandomWalk` base class and all proposals that are derived from it.
"""

class GaussianRandomWalk(Proposal):
    
    """
    Generates proposals using a Gaussian random walk.

    Attributes
    ---------
    sqrt_cov : torch.Tensor
        Lower cholesky decomposition of the covariance of Gaussian proposal
    dim : int
        Dimension of the Gaussian proposal
    """

    def __init__(
            self,
            sqrt_cov : torch.Tensor = None,
            cov : torch.Tensor = None
    ):
        
        """
        GaussianRandomWalk constructor.

        Parameters
        ----------
        sqrt_cov : torch.Tensor
            Lower cholesky decomposition of the covariance of Gaussian proposal. Must be 2-dimensional and square.
        cov : torch.Tensor
            Covariance matrix of the Gaussian proposal. Must be symmetric and positive-definite.
        """
        
        if sqrt_cov is not None:
            if sqrt_cov.ndim !=2 or sqrt_cov.shape[0] != sqrt_cov.shape[1]:
                raise ValueError(f'sqrt_cov must be a square matrix but instead it has shape {sqrt_cov.shape}')
            self.sqrt_cov = sqrt_cov
        elif cov is not None:
            try:
                self.sqrt_cov = torch.linalg.cholesky(cov, upper=False)
            except:
                raise ValueError('cov must be a symmetric positive-definite matrix')
        else:
            raise ValueError('GaussianRandomWalk requires either sqrt_cov or cov for initialization')

        self.dim = self.sqrt_cov.shape[0]
        
    @property
    def is_symmetric(
            self
    ) -> bool:
        
        """
        Getter of `is_symmetric` property. 

        Returns
        ----------
        bool
            Gaussian distributions are symmetric, so this always returns `True`
        """
        
        return True


    def propose(
            self,
            y : Tuple[Sample, ...],
            N : int = 1
    ) -> Union[Sample, Tuple[Sample, ...]]:
        
        """
        Draws a sample from the Gaussian random walk

        Parameters
        ----------
        y : Tuple[Sample, ...]
            The Gaussian distribution centers itself around the first Sample in this list
        N : int
            Number of samples to draw. Default is 1

        Returns
        ----------
        Union[Sample, Tuple[Sample, ...]]
            The samples drawn from the Gaussian random walk.
            If `N` is 1, a Sample object is returned
            If `N`>1, a list of Sample objects is returned
        """
        
        sample_points = torch.randn(N, self.dim) @ self.sqrt_cov.T + y[0].point
        return [Sample(sample_points[ii]) for ii in range(N)] if N > 1 else Sample(sample_points)

    def log_prob(
            self,
            y : Tuple[Sample, ...]
    ) -> torch.Tensor:
        
        """
        Evaluates the log probability density of the Gaussian distribution

        Parameters
        ----------
        y : Tuple[Sample, ...]
            List of Samples representing the current state and all proposals for a given iteration ordered from first proposal tier to current proposal tier

        Returns
        ----------
        torch.Tensor
            Log probability density for a Gaussian centered at the first Sample in `y` evaluated at the last Sample in `y`.
        """
        
        return -0.5 * torch.sum(torch.linalg.solve_triangular(self.sqrt_cov, (y[0].point - y[-1].point).T, upper=False)**2, dim=0)


class AdaptiveCovariance(GaussianRandomWalk):

    """
    A Gaussian random walk with covariance that adapts according to the sample covariance. This is a child class of the `GaussianRandomWalk` class.

    Attributes
    ---------
    sqrt_cov : torch.Tensor
        Lower cholesky decomposition of the covariance of Gaussian proposal
    cov : torch.Tensor
        Covariance matrix of the Gaussian proposal
    dim : int
        Dimension of the Gaussian proposal
    sd : float
        Scaling factor of the proposal covariance.
    n0 : int
        Number of Markov chain iterations before using adaptive covariance
    eps : float
        Small positive value to add to covariance diagonal during adaptation to encourage positive-definiteness
    iteration : int
        The iteration number.
    mean : torch.Tensor
        The running sample mean
    """

    def __init__(
            self,
            cov : torch.Tensor,
            sd : float = None,
            n0 : int = 200,
            eps : float = 1e-8
    ):
        
        """
        AdaptiveCovariance constructor.

        Parameters
        ----------
        cov : torch.Tensor
            Covariance matrix of the Gaussian proposal. Must be 2-dimensional and square.
        sd : float
            Scaling factor of proposal covariance. If `None`, the default value is $2.4^2 / d$, where $d$ is the proposal dimension.
        n0 : int
            Number of Markov chain iterations before using adaptive covariance. Default is 200
        eps : float
            Small positive value to be added to the covariance diagonal during adaptation to encourage positive-definiteness. Default is 1e-8
        """

        super().__init__(cov = cov)
        self.sd = 2.4**2 / self.dim if sd is None else sd
        self.sqrt_cov *= sqrt(self.sd)
        self.cov = cov * self.sd
        self.n0 = n0
        self.eps = eps
        
        self.iteration = 0
        self.mean = torch.zeros(self.dim)

    @property
    def cov(
            self
    ) -> torch.Tensor:
        
        """
        Getter for the `cov` property.

        Returns
        ----------
        torch.Tensor
            If `iteration`>`n0`, the sample covariance is returned. Otherwise, the initial covariance is returned.
        """
        
        if self.iteration > self.n0:
            return self._cov_ii
        else:
            return self._cov_0

    @cov.setter
    def cov(
        self,
        value : torch.Tensor
    ):
        
        """
        Setter for the `cov` property

        Parameters
        ----------
        value : torch.Tensor
            A covariance matrix.
        """
        
        self._cov_0 = value
        self._cov_ii = value.clone()


    def _adapt(
            self,
            x : torch.Tensor
    ):
        
        """
        Adaptation algorithm for updating the adaptive covariance. Updates the sample mean `mean` and covariance `_cov_ii`

        Parameters
        ----------
        x : torch.Tensor
            The most recent state value in the Markov chain.
        """
        
        self.iteration += 1

        self._cov_ii *= ((self.iteration - 1) / self.iteration)
        outer = self.iteration * torch.outer(self.mean, self.mean) + torch.outer(x, x)
        
        self.mean = (self.mean * (self.iteration - 1) + x) / self.iteration
        
        outer_mean = (self.iteration + 1) * torch.outer(self.mean, self.mean)

        self._cov_ii += (self.sd / self.iteration) * (outer - outer_mean + self.eps * torch.eye(self.dim))

        if self.iteration > self.n0:
            self.sqrt_cov = torch.linalg.cholesky(self._cov_ii, upper=False)

    def propose(
            self,
            y : Tuple[Sample, ...],
            N : int = 1
    ) -> Union[Sample, Tuple[Sample, ...]]:
        
        """
        Generates a proposed state from the Gaussian random walk with adaptive covariance.

        Parameters
        ----------
        y : Tuple[Sample, ...]
            A list of Sample objects containing the current state and proposals from each proposal tier
        N : int
            Number of proposed states to draw. Default is 1

        Returns
        ----------
        Union[Sample, Tuple[Sample, ...]]
            Returns the proposed state wrapped inside a Sample object if `N` is 1. Otherwise, returns a list of Sample objects, each containing a proposed state
        """
        
        return super().propose(y,N)


class ScaledCovariance(GaussianRandomWalk):

    """
    Wraps a GaussianRandomWalk
    """

    def __init__(
            self,
            covariance : GaussianRandomWalk,
            gamma : float = 1e-2
    ):
        
        """
        ScaledCovariance constructor.

        Parameters
        ----------
        covariance : GaussianRandomWalk
            The Gaussian random walk proposal to be scaled.
        gamma : float
            The scaling factor for the second-stage proposal covariance. Default is 1e-2.
        """
        
        self.covariance = covariance
        self.sqrt_gamma = sqrt(gamma)

    @property
    def sqrt_cov(
            self
    ) -> torch.Tensor:
        
        """
        Getter for the `sqrt_cov` property. 

        Returns
        ----------
        torch.Tensor
            The `sqrt_cov` property of the wrapped GaussianRandomWalk object scaled by `sqrt_gamma`.
        """
        
        return self.covariance.sqrt_cov * self.sqrt_gamma

    @property
    def dim(
            self
    ) -> int:
        
        """
        Getter for the `dim` property. 

        Returns
        ----------
        int
            The `dim` property of the wrapped GaussianRandomWalk object.
        """
        
        return self.covariance.dim



"""mcmc-samplers: MCMC sampling in PyTorch"""

# Samples.
from mcmc_samplers.samples.sample import Sample
from mcmc_samplers.samples.hamiltonian import HamiltonianSample

# Proposals.
from mcmc_samplers.proposals.proposal import Proposal
from mcmc_samplers.proposals.gaussian_random_walk import GaussianRandomWalk
from mcmc_samplers.proposals.gaussian_random_walk import AdaptiveCovariance
from mcmc_samplers.proposals.gaussian_random_walk import ScaledCovariance
from mcmc_samplers.proposals.hamiltonian import HamiltonianDynamics

# Samplers.
from mcmc_samplers.samplers.sampler import Sampler
from mcmc_samplers.samplers.delayed_rejection import DelayedRejection
from mcmc_samplers.samplers.delayed_rejection import DelayedRejectionAdaptiveMetropolis
from mcmc_samplers.samplers.delayed_rejection import MetropolisHastings
from mcmc_samplers.samplers.hamiltonian import HamiltonianMonteCarlo

# Visualization
from mcmc_samplers.visualization.sample_visualizer import SampleVisualizer

__all__ = (
    "AdaptiveCovariance",
    "DelayedRejection",
    "DelayedRejectionAdaptiveMetropolis",
    "GaussianRandomWalk",
    "HamiltonianDynamics",
    "HamiltonianMonteCarlo",
    "HamiltonianSample",
    "MetropolisHastings",
    "Proposal",
    "Sample",
    "Sampler",
    "SampleVisualizer",
    "ScaledCovariance"
)




import torch
from torch.distributions.multivariate_normal import MultivariateNormal
from mcmc_samplers import *

import matplotlib.pyplot as plt

"""
Toy example to demonstrate the functionality of the DRAM and HMC samplers.
"""

class Banana:

    """
    2D dimensional target distribution with a banana-shaped joint distribution.

    Attributes
        ----------
        mvn : MultivariateNormal
            Multivariate normal distribution that will be used to evaluate the log probability.
    """

    def __init__(
            self
    ):

        """
        Banana constructor.
        """
        
        mean = torch.zeros(2)
        cov = torch.tensor([[1., 0.9], [0.9, 1.]])
        self.mvn = MultivariateNormal(mean, cov)

    def log_prob(
            self,
            x : torch.Tensor
    ) -> torch.Tensor:

        """
        Evaluates the log probability of the banana distribution.

        Parameters
        ----------
        x : torch.Tensor
            Point at which to evaluate the log probability density.

        Returns
        ----------
        torch.Tensor
            The log probability density of the banana distribution at the point `x`.
        """
        
        x = torch.atleast_2d(x)
        y = torch.cat((x[:,0:1], x[:,1:2] + (x[:,0:1] + 1)**2), dim=1)
        return self.mvn.log_prob(y)




if __name__ == "__main__":

    torch.manual_seed(0) # for reproducibility
    
    target = Banana()

    init_sample = torch.tensor([0.,-1.])
    init_cov = torch.tensor([[1., 0.9], [0.9, 1.]])

    num_samples = int(1e4)


    dram = DelayedRejectionAdaptiveMetropolis(
        target = target.log_prob,
        x0 = init_sample,
        cov = init_cov
    )
    
    hmc = HamiltonianMonteCarlo(
        target = target.log_prob,
        x0 = torch.nn.Parameter(init_sample),
        step_size = 0.2,
        num_steps = 3
    )

    samplers = [dram, hmc]
    titles = ['DRAM', 'HMC']
    labels = ['$x_1$', '$x_2$']
    visualizers = [None, None]

    for ii, sampler in enumerate(samplers):
        samples, log_probs = sampler(num_samples)
        print(f'{titles[ii]} acceptence rate: {100*sampler.acceptance_ratio:.2f}%')
    
        visualizers[ii] = SampleVisualizer(samples)
        visualizers[ii].triangular_hist(bins=50,
                                   labels=labels,
                                   titles=[titles[ii]])
        visualizers[ii].chains(labels=labels,
                               titles=[titles[ii]])

    plt.show()



import torch
from mcmc_samplers import Sample
from abc import ABC, abstractmethod
from typing import Tuple

"""
Defines the Proposal base class. Proposal objects propose states to which to move an MCMC sampler. Proposal objects are also able to evaluate the log probability of their underlying proposal distribution.
"""

class Proposal(ABC):
    
    """
    Abstract base class of the Proposal class.

    Attributes
    ----------
    is_symmetric : bool
        True if the proposal is symmetric and false otherwise.
    """

    @abstractmethod
    def __init__(
            self
    ):
        
        """
        Proposal constructor.
        """
        
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_symmetric(
            self
    ) -> bool:

        """
        Getter of the property `is_symmetric``.
        """
        
        raise NotImplementedError()

    @abstractmethod
    def propose(
            self,
            y : Tuple[Sample, ...]
    ) -> Sample:
        
        """
        Generates a sample from the proposal distribution.
        """
        
        raise NotImplementedError()

    @abstractmethod
    def log_prob(
            self,
            y : Tuple[Sample, ...]
    ) -> torch.Tensor:
        
        """
        Evaluates the log of unnormalized probability density function
        """
        
        raise NotImplementedError()



import torch
from mcmc_samplers import Proposal, Sample
from abc import ABC
from collections.abc import Callable
from typing import Union, Tuple

"""
Defines the Sampler abstract base class. Sampler objects are used to execute MCMC algorithms.
"""

class Sampler(ABC):
    
    """
    Abstract base class of the Sampler class.

    Attributes
    ----------
    target : Callable[[torch.Tensor], [torch.Tensor]]
        Function that evaluates the log probability density of a target distribution.
    x : Sample
        Sample representing the current position of the Markov chain
    proposals : Tuple[Proposal, ...]
        List of proposals.
    acceptance_kernels : Tuple[Callable[[Tuple[Sample, ...]],[torch.Tensor]], ...]
        List of acceptance kernels. 
    """

    def __init__(
            self,
            target : Callable[[torch.Tensor], [torch.Tensor]],
            x0 : Union[Sample, torch.Tensor],
            proposals : Tuple[Proposal, ...],
            acceptance_kernels : Tuple[Callable[[Sample, Tuple[Sample, ...]],[torch.Tensor]], ...]
    ):
        
        """
        Sampler constructor.

        Parameters
        ----------
        target : Callable[[torch.Tensor], [torch.Tensor]]
            evaluates the log probability of the target distribution.
        x0 : Union[Sample, torch.Tensor]
            point at which to initialize the Markov chain
        proposals : Tuple[Proposal, ...]
            list of proposals. 
        acceptance_kernels : Tuple[Callable[[Sample, Tuple[Sample, ...]],[torch.Tensor]], ...]
            list of acceptance kernels. Each acceptance kernel evaluates the log acceptance probability of a proposed sample given the current position of the Markov chain (and possibly given rejected proposals at earlier stages)
        """

        self.acceptance_ratio = 0
        self.target = target
        self.x = torch.atleast_2d(x0)
        self.proposals = proposals
        self.acceptance_kernels = acceptance_kernels

    @property
    def x(
            self
    ) -> Sample:
        
        """
        Getter for the property `x`

        Returns
        ----------
        Sample
            the current position of the Markov chain.
        """
        
        return self._x

    @x.setter
    def x(
            self,
            value : Union[Sample, torch.Tensor]
    ):
        
        """
        Setter for the property `x`.

        Parameters
        ----------
        value : Union[Sample, torch.Tensor]
            Markov chain state at which to position the sampler.
        """
        
        if isinstance(value, Sample):
            self._x = value
            if value.log_prob is None:
                self._x.log_prob = self.target(value.point)
        elif isinstance(value, torch.Tensor):
            self._x = Sample(value, self.target(value))
        else:
            raise ValueError(f'x must be of type either Sample or torch.Tensor')

    def _sample(
            self
    ) -> int:
        
        """
        Advances the Markov chain forward one step

        Returns
        ----------
        int
            1 if a proposed sample is accepted. In this case the sampler moves its position to the accepted point
            0 if all proposed samples are rejected. In this case, the sampler remains in its current state.
        """
        
        y = [self.x]
        for proposal, acceptance_prob in zip(self.proposals, self.acceptance_kernels):
            y.append(proposal.propose(y))
            alpha = acceptance_prob(y)
            if alpha > torch.log(torch.rand(1)): # accept
                self.x = y[-1]
                return 1
        return 0 # proposal rejected at every stage

    def __call__(
            self,
            N : int = 1
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        
        """
        Runs the sampler for as long as specified.

        Parameters
        ----------
        N : int
            Number of steps to advance the sampler. Default is 1

        Returns
        ----------
        Tuple[torch.Tensor, torch.Tensor]
            The first return object is a tensor representing the Markov chain over `N` iterations
            The second return object is a tensor holding the `target` log probability values of each state in the Markov chain.
        """

        samples = torch.zeros(N, self.x.point.shape[1])
        log_probs = torch.zeros(N)
        for ii in range(N):
            self.acceptance_ratio = (ii * self.acceptance_ratio + self._sample()) / (ii+1)
            samples[ii] = self.x.point.detach()
            log_probs[ii] = self.x.log_prob
            
        return samples, log_probs





import torch
import numpy as np
from typing import Tuple
import matplotlib.pyplot as plt

"""
Plotting functions for visualizing sample distributions.
"""

class SampleVisualizer:

    """
    Class to hold and manipulate samples for plotting.

    Attributes
    ----------
    samples : torch.Tensor
        Samples from target distribution.
    dim : int
        Dimension of samples.
    """

    def __init__(
            self,
            samples : torch.Tensor
    ):

        """
        Sample_Visualizer constructor.

        Parameters
        ----------
        samples : torch.Tensor
            2D tensor holding samples from target distribution. First dimension should be number of samples and second dimension should be sample dimension.
        """

        self.samples = samples.detach()
        self.dim = samples.shape[1]

    def _get_defaults(
            self,
            labels : Tuple[str, ...],
            titles : Tuple[str, ...],
            partition : Tuple[Tuple[int, ...], ...]
    ):
        if partition is None:
            partition = [range(self.dim)]
        if labels is None:
            labels = [f'$\\theta_{ii}$' for ii in range(self.dim)]
        if titles is None:
            titles = [None for ii in range(len(partition))]
        else:
            assert len(titles) == len(partition)
        return labels, titles, partition
        

    def chains(
            self,
            step : int = 1,
            labels : Tuple[str, ...] = None,
            titles : Tuple[str, ...] = None,
            partition : Tuple[Tuple[int, ...], ...] = None,
    ) -> Tuple[Tuple[plt.Figure, ...], Tuple[np.ndarray, ...]]:

        """
        Plots the sample chains.

        Parameters
        ----------
        step : int
            Step size by which to subsample the Markov chain. Default is 1, which results in no subsampling.
        labels : Tuple[str, ...]
            Iterable containing a label for each sample component. The length should be equal to the sample dimension. By default, the $i$th sample is labeled as $\theta_i$.
        titles : Tuple[str, ...]
           Iterable containing a suptitle for each figure. By default, empty strings are used for figure titles.
        partition : Tuple[Tuple[int, ...], ...]
            Iterable containing iterables of indices. Each iterable produces a plot containing the samples that correspond to the indices within that iterable. The indices refer to the sample dimension and not the chain iteration. By default, all samples are plotted in a single plot by setting `partition=[range(self.dim)]'

        Returns
        ----------
        Tuple[Tuple[plt.Figure, ...], Tuple[np.ndarray, ...]]
            The first Tuple is a list of Figure objects corresponding to each plot. The second Tuple is a list numpy arrays of the Axes objects from each plot. The length of each list equals the length of `partition`.
        """

        labels, titles, partition = self._get_defaults(labels,titles,partition)

        figs = []
        axes = []
        xx = range(0, self.samples.shape[0], step)

        for ii, idx in enumerate(partition):
            fig, axs = plt.subplots(len(idx), sharex=True)
            for jj, idx_jj in enumerate(idx):
                axs[jj].plot(xx, self.samples[::step, idx_jj])
                axs[jj].set_ylabel(labels[idx_jj])

            axs[-1].set_xlabel('Iteration')
            fig.suptitle(titles[ii])
            figs.append(fig)
            axes.append(axs)

        return figs, axes



    def triangular_hist(
            self,
            bins : int = None,
            step : int = 1,
            labels : Tuple[str, ...] = None,
            titles : Tuple[str, ...] = None,
            partition : Tuple[Tuple[int, ...], ...] = None,
    ) -> Tuple[Tuple[plt.Figure, ...], Tuple[np.ndarray, ...]]:
        
        """
        Plots 1D and 2D histograms of the samples.

        Parameters
        ----------
        bins : int
            Number of bins to use along each axis in each histogram. Default is 10.
        step : int
            Step size by which to subsample the Markov chain. Default is 1, which results in no subsampling.
        labels : Tuple[str, ...]
            Iterable containing a label for each sample component. The length should be equal to the sample dimension. By default, the $i$th sample is labeled as $\theta_i$.
        titles : Tuple[str, ...]
           Iterable containing a suptitle for each figure. By default, empty strings are used for figure titles.
        partition : Tuple[Tuple[int, ...], ...]
            Iterable containing iterables of indices. Each iterable produces a plot containing the samples that correspond to the indices within that iterable. The indices refer to the sample dimension and not the chain iteration. By default, all samples are plotted in a single plot by setting `partition=[range(self.dim)]'

        Returns
        ----------
        Tuple[Tuple[plt.Figure, ...], Tuple[np.ndarray, ...]]
            The first Tuple is a list of Figure objects corresponding to each plot. The second Tuple is a list numpy arrays of the Axes objects from each plot. The length of each list equals the length of `partition`.
        """

        labels, titles, partition = self._get_defaults(labels,titles,partition)

        figs = []
        axes = []

        for ii, idx in enumerate(partition):
            len_idx = len(idx)
            fig, axs = plt.subplots(len_idx, len_idx, squeeze=False)
            for row in range(len_idx):
                for col in range(len_idx):
                    # upper triangle. no plots.
                    if row < col:
                        axs[row,col].axis('off')
                    # diagonal. 1D histograms.
                    elif row == col:
                        axs[row,col].hist(self.samples[::step,idx[row]], bins=bins)
                        plt.setp(axs[row,col].get_yaxis(), visible=False)        
                    # lower triangle. 2D histograms.
                    else:
                        # share x-axis with 1D histogram from same column.
                        axs[row-1,col].get_shared_x_axes().join(axs[row-1,col],axs[row,col])
                        # share y-axis with 2D histograms from same row.
                        if col > 0:
                            axs[row,col-1].get_shared_y_axes().join(axs[row,col-1],axs[row,col])
                        
                        axs[row,col].hist2d(self.samples[::step,idx[col]], self.samples[::step,idx[row]], bins=bins)

                    # Adjust axis labels and axis visibility.
                    if col == 0 and row > 0:
                        axs[row,col].set_ylabel(labels[idx[row]])
                    else:
                        plt.setp(axs[row,col].get_yticklabels(), visible=False)        
                    if row == len(idx)-1:
                        axs[row,col].set_xlabel(labels[idx[col]])
                    else:
                        plt.setp(axs[row,col].get_xticklabels(), visible=False)

            fig.suptitle(titles[ii])
            figs.append(fig)
            axes.append(axs)
            
        return figs, axes



import torch

"""
Base class used throughout the 'mcmc-samplers' package
"""

class Sample:

    """
    Data structure representing an MCMC sample.

    Attributes
    ----------
    point : torch.Tensor
        The sample value
    log_prob : torch.Tensor
        The log probability density of a distribution of interest evaluated at the sample value `point`. The distribution of interest is typically the target distribution of the MCMC algorithm.
    """

    def __init__(
            self,
            point : torch.Tensor = None,
            log_prob : torch.Tensor = None
    ):
        
        """
        Sample constructor

        Parameters
        ----------
        point : torch.Tensor
            sample value
        log_prob : torch.Tensor, optional
            log probability density of a distribution of interest evaluated at the sample value `point`.
        """

        self.point = point
        self.log_prob = log_prob

    @property
    def shape(
            self
    ):
        return self.point.shape

    @shape.setter
    def shape(
            self,
            value
    ):
        raise AttributeError("attribute 'shape' of 'mcmc-samplers.Sample' objects is not writable")

    def __str__(
            self
    ) -> str:
        string = f'sample(point={self.point})'
        if self.log_prob is not None:
            string += f', log_prob={self.log_prob}'
        return string

    def __len__(
            self
    ) -> int:
        return len(self.point)



import torch
from torch.nn import Parameter
from mcmc_samplers import DelayedRejection, HamiltonianDynamics, HamiltonianSample, Sample
from typing import Union
from collections.abc import Callable

"""
Defines the HamiltonianMonteCarlo class as a child of the `DelayedRejection` base class
"""

class HamiltonianMonteCarlo(DelayedRejection):

    """
    Implementation of the Hamiltonian Monte Carlo or Hybrid Monte Carlo (HMC) algorithm.

    Citation:
        Neal, Radford M. "MCMC using Hamiltonian dynamics." Handbook of Markov chain Monte Carlo 2.11 (2011): 2.

    Attributes
    ----------
    target : Callable[[torch.Tensor], [torch.Tensor]],
        Function that evaluates the log probability of the target distribution.
    x0 : HamiltonianSample
        Initial point of the Markov chain.
    dim : int
        Dimension of the sampler.
    proposals : Tuple[Proposal]
        HamiltonianDynamics proposal.
    acceptance_kernels : Tuple[Callable[[Tuple[Sample, ...]],[torch.Tensor]]]
        The Metropolis-Hastings acceptance kernel. In this case, it is the ratio of Hamiltonians of the current and proposed points.
    """

    def __init__(
            self,
            target : Callable[[torch.Tensor], [torch.Tensor]],
            x0 : Union[Sample, torch.Tensor],
            step_size : float = 1e-1,
            num_steps : int = 1
    ):

        """
        HamiltonianMonteCarlo constructor.

        Parameters
        ----------
        target : Callable[[torch.Tensor], [torch.Tensor]]
        Function that evaluates the log probability density of a target distribution
        x0 : Union[Sample, torch.Tensor]
            Initial point of the Markov chain for the position variables.
        step_size : float
            Step size to integrate the Hamiltonian dynamics. Default is 1e-1.
        num_steps : int
            Number of steps to integrate the Hamiltonian dynamics. Default is 1.
        """

        self.dim = len(x0)
        super().__init__(
            target = target,
            x0 = x0,
            proposals = [
                HamiltonianDynamics(
                    target,
                    self.dim,
                    step_size,
                    num_steps
                )
            ]
        )


    @property
    def x(
            self
    ) -> HamiltonianSample:

        """
        Getter for the `x` property.

        Returns
        ----------
        HamiltonianSample
            Current position and momentum of the sampler
        """
        
        return self._x

    @x.setter
    def x(
            self,
            value : Union[torch.Tensor, Sample]
    ):

        """
        Setter for the `x` property. Initializes the position variable as a torch.nn.Parameter. Then it evaluates the `self.target` and calls `backward` to compute the gradient.

        Parameters
        ----------
        value : Union[torch.Tensor, Sample]
           Either a torch.Tensor or Sample representing the position variable of the Hamiltonian system.
        """

        if isinstance(value, HamiltonianSample):
            self._x = HamiltonianSample(value.q, value.p)
        else:
            self._x = HamiltonianSample(value)

        # compute gradient    
        self._x.log_prob = self.target(self._x.point)
        self._x.log_prob.backward()
        



import torch
from mcmc_samplers import *
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Union, Tuple

"""
Defines the `DelayedRejection` abstract base class and all MCMC samplers that are derived from it.
"""

class DelayedRejection(Sampler, ABC):
    
    """
    The `DelayedRejection` abstract base class. It is a child class of the `Sampler` class. This class executes the multi-tiered proposal algorithm known as ``delayed rejection.''

    Citation:
        Green, Peter J., and Antonietta Mira. ``Delayed rejection in reversible jump Metropolis-Hastings.'' Biometrika 88.4 (2001): 1035-1053.

    Attributes
    ----------
    Carries the same attributes as the parent class `Sampler`.
    """

    def __init__(
            self,
            target : Callable[[torch.Tensor], [torch.Tensor]],
            x0 : Union[Sample, torch.Tensor],
            proposals : Tuple[Proposal, ...],
            num_stages : int = 1
    ):
        
        """
        `DelayedRejection` constructor. Constructs the acceptance_kernels according to the delayed rejection algorithm.

        Parameters
        ----------
        target : Callable[[torch.Tensor], [torch.Tensor]]
            evaluates the log probability of the target distribution.
        x0 : Union[Sample, torch.Tensor]
            point at which to initialize the Markov chain
        proposals : Tuple[Proposal, ...]
            list of proposals.
        num_stages : int
            Number of proposal stages/tiers. Default is 1
        """
        
        super().__init__(
            target = target,
            x0 = x0,
            proposals = proposals,
            acceptance_kernels = [lambda y, stage=ii : self._metropolis_hastings(y, stage) for ii in range(1, num_stages+1)]
        )

        self._denom = None
        self._alpha = None

    def _metropolis_hastings(
            self,
            y : Tuple[Sample, ...],
            stage : int = 1
    ) -> torch.Tensor:
        
        """
        Evaluates the log acceptance probability of a proposed sample at a given stage level.

        Parameters
        ----------
        y : Tuple[Sample, ...]
           List containing samples representing the current state of the sampler, the proposed sample, and all rejected proposals, if any.  The ordering of the list is as follows: [x, y_1, ... , y_stage]
        stage : int
            The stage/tier of the proposal. Default is 1 representing the standard Metropolis-Hastings acceptance probability

        Returns
        ----------
        int
            the log acceptance probability
        """

        assert len(y) == stage + 1
        assert self._denom is not None
        
        if stage > 1:
            assert self._alpha is not None
        
        if y[-1].log_prob is None:
            y[-1].log_prob = self.target(y[-1].point)

        denom = self._denom.clone()
        if stage > 1:
            denom += torch.log(1 - torch.exp(self._alpha))
            if self.proposals[stage-2].is_symmetric:
                denom += self.proposals[stage-2].log_prob(y[:-1])

        self._denom = y[-1].log_prob.clone()
        for ii in range(1,stage):
            self.acceptance_kernels[ii-1](y[:-(ii+2):-1])

        numer = self._denom.clone()
        if stage > 1:
            numer += torch.log(1 - torch.exp(self._alpha))
            if self.proposals[stage-2].is_symmetric:
                numer += self.proposals[stage-2].log_prob(y[:0:-1])

        if not self.proposals[stage-1].is_symmetric:
            numer += self.proposals[stage-1].log_prob(y[::-1])
            denom += self.proposals[stage-1].log_prob(y)

        self._denom = denom
        self._alpha = min(torch.zeros(1), numer - self._denom)
        return self._alpha

    def _sample(
            self
    ) -> int:
        
        """
        Advances the Markov chain one step. 

        Returns
        ----------
        int
            1 if a proposed sample is accepted and 0 otherwise.
        """
        
        self._denom = self.x.log_prob
        return super()._sample()

    
class DelayedRejectionAdaptiveMetropolis(DelayedRejection):

    """
    Implementation of the delayed rejection adaptive Metropolis (DRAM) sampler. This is a child class of the `DelayedRejection`

    Citation:
        Haario, Heikki, et al. "DRAM: efficient adaptive MCMC." Statistics and computing 16 (2006): 339-354.

    Attributes
    ----------
    The attributes are the same as those in the `DelayedRejection` parent class.    
    """

    def __init__(
            self,
            target : Callable[[torch.Tensor], [torch.Tensor]],
            x0 : Union[Sample, torch.Tensor],
            cov : torch.Tensor = None,
            sd : float = None,
            gamma : float = 1e-2,
            n0 : int = 200,
            eps : float = 1e-8
    ):
        
        """
        DelayedRejectionAdaptiveMetropolis constructor.

        target : Callable[[torch.Tensor], [torch.Tensor]]
            function evaluating the log probability of the target distribution
        x0 : Union[Sample, torch.Tensor]
            initial state of the Markov chain
        cov : torch.Tensor, optional
            Tensor representing the proposal covariance
        sd : float, optional
            Scaling factor applied to the proposal covariance. By default, the value is set to be $2.4^2 / d$, where $d$ is the dimension of the proposal
        gamma : float
            Scaling factor applied to the first-stage proposal covariance to yield the second-stage proposal covariance. The default is 1e-2
        n0 : int
            Number of iterations to run the sampler before using the empirical sample covariance as the first-stage proposal covariance. Default is 200
        eps : float
            Small positive value added to the diagonal of the proposal covariance during adaptation to encourage positive-definitess. Default is 1e-8
        """
        
        proposals = [AdaptiveCovariance(cov, sd, n0, eps)]
        proposals.append(ScaledCovariance(proposals[0], gamma))
        super().__init__(
            target = target,
            x0 = x0,
            proposals = proposals,
            num_stages = 2
        )

    def _sample(
            self
    ) -> int:
        
        """
        Advances the Markov chain one step. First, this method adapts the proposal covariance with the adaptive Metropolis (AM) algorithm. Then it samples with the delayed rejection (DR) algorithm.

        Returns
        ----------
        int
            1 if a proposed sample is accepted and 0 otherwise.
        """
        
        self.proposals[0]._adapt(self.x.point.squeeze())
        return super()._sample()

    

class MetropolisHastings(DelayedRejection):

    """
    The basic Metropolis-Hastings MCMC sampler. Samples are proposed with a Gaussian random walk and accepted with probability given by the Metropolis-Hastings acceptance kernel. This is a child of the `DelayedRejection` class.

    Attributes
    ----------
    The attributes are the same as those of the `DelayedRejection` parent class.
    """

    def __init__(
            self,
            target : Callable[[torch.Tensor], [torch.Tensor]],
            sqrt_cov : torch.Tensor = None,
            cov : torch.Tensor = None
    ):
        
        """
        MetropolisHastings constructor

        target : Callable[[torch.Tensor], [torch.Tensor]]
            function evaluating the log probability of the target distribution
        sqrt_cov : torch.Tensor
            the lower cholesky decomposition of the covariance of the Gaussian random walk proposal
        cov : torch.Tensor
            the covariance of the Gaussian random walk proposal. Do not need to pass this if `sqrt_cov` is already passed.
        """
        
        super().__init__(
            target = target,
            proposals = [GaussianRandomWalk(sqrt_cov, cov)]
        )



import torch
from torch.nn import Parameter
from mcmc_samplers import Proposal, HamiltonianSample, Sample
from torch.distributions.multivariate_normal import MultivariateNormal
from typing import Tuple
from collections.abc import Callable

"""
Defines the HamiltonianDynamics class as a child of the `Proposal` base class.
"""

class HamiltonianDynamics(Proposal):
    
    '''
    Proposal using Hamiltonian dynamics.

    Attributes
    ----------
    target : Callable[[torch.Tensor], [torch.Tensor]]
        Function that evaluates the log probability of the target distribution. In the Hamiltonian dyanmics, this represents the negative potential energy.
    dim : int
        Dimension of the sampler.
    step_size : float
        Step size to use to integrate the Hamiltonian dynamics.
    num_steps : int
        Number of steps to integrate the Hamiltonian dynamics.
    kinetic_energy : torch.Distribution.MultivariateNormal
        The negative log of this distribution represents the kinetic energy.
    '''
    
    def __init__(
            self,
            target : Callable[[torch.Tensor], [torch.Tensor]],
            dim : int,
            step_size : float = 1e-1,
            num_steps : int = 1
    ):

        """
        Description

        Parameters
        ----------
        target : Callable[[torch.Tensor], [torch.Tensor]]
            Function that evaluates the log probability of the target distribution. In the Hamiltonian dyanmics, this represents the negative potential energy.
        dim : int
            Dimension of the sampler.
        step_size : float
            Step size to use to integrate the Hamiltonian dynamics. Default is 1e-1.
        num_steps : int
            Number of steps to integrate the Hamiltonian dynamics. Default is 1.
        """

        self.target = target
        self.dim = dim
        self.step_size = step_size
        self.num_steps = num_steps

        self.kinetic_energy = MultivariateNormal(
            loc = torch.zeros(self.dim),
            covariance_matrix = torch.eye(self.dim)
        )

    @property
    def is_symmetric(
            self
    ) -> bool:

        """
        Getter for property `is_symmetric`. 

        Returns
        ----------
        bool
            The Hamiltonian dynamics proposal is not symmetric, so value is always False.
        """
        
        return False

    def _leapfrog(
            self,
            y : HamiltonianSample
    ) -> HamiltonianSample:

        """
        Leapfrog integrator.

        Parameters
        ----------
        y : HamiltonianSample
            The current position and momentum of the sampler stored as a HamiltonianSample object.

        Returns
        ----------
        HamiltonianSample
            The position and momentum approximately integrated `self.step_size * self.num_steps` forward in time and stored in a HamiltonianSample object.
        """

        q = Sample(y.point.detach().clone())
        p = y.p.clone()

        # Integrate for `self.num_steps` steps
        p = p + (self.step_size / 2) * y.point.grad
        for ii in range(self.num_steps):
            q.point = Parameter(q.point + self.step_size * p)
            q.log_prob = self.target(q.point)
            q.log_prob.backward()
            if ii != self.num_steps - 1:
                p = p + self.step_size * q.point.grad
        p = p + (self.step_size / 2) * q.point.grad
        
        return HamiltonianSample(q, p)

    def propose(
            self,
            y : Tuple[HamiltonianSample, ...]
    ) -> HamiltonianSample:

        """
        Proposes a sample using Hamiltonian dynamics.

        Parameters
        ----------
        y : Tuple[HamiltonianSample, ...]
            List of current and proposed HamiltonianSample objects

        Returns
        ----------
        HamiltonianSample
            Proposed sample consisting of a position and momentum variable.
        """

        y[0].p = self.kinetic_energy.sample()
        return self._leapfrog(y[0])

    def log_prob(
            self,
            y : Tuple[HamiltonianSample, ...] = None
    ) -> torch.Tensor:

        """
        Evaluates the log probability of the proposal kernel.

        Parameters
        ----------
        y : Tuple[HamiltonianSample, ...]
            List of current and proposed HamiltonianSample objects.

        Returns
        ----------
        torch.Tensor
            Log probability density of the proposal. Equivalent to the negative kinetic energy function.
        """

        return self.kinetic_energy.log_prob(y[-1].p)



import torch
from torch.nn import Parameter
from mcmc_samplers import Sample
from typing import Union

"""
Defines the HamiltonianSample class as a child of the `Sample` base class.
"""

class HamiltonianSample(Sample):

    """
    A child of the Sample class for Hamiltonian Monte Carlo (HMC). Objects belonging to this class carry a position and momentum variable.

    Attributes
    ----------
    q : Sample
        The position variable. The point value is initialized as torch.nn.Parameter so that gradients with resepct to `q` can be computed with `backward`.
    p : torch.Tensor
        The momentum variable. Auxiliary variable in HMC.
    """

    def __init__(
            self,
            q = Union[Sample, torch.Tensor],
            p : torch.Tensor = None
    ):

        """
        HamiltonianSample constructor.

        Parameters
        ----------
        q : Union[Sample, torch.Tensor]
            The position of the Hamiltonian dynamics. Either a Sample or torch.Tensor can be passed.
        p : torch.Tensor, optional
            The momentum of the Hamiltonian dynamics.
        """
        
        if isinstance(q, Sample):
            self.q = Sample(Parameter(q.point))
            self.log_prob = q.log_prob
        elif isinstance(q, torch.Tensor):
            self.q = Sample(Parameter(q))
        else:
            raise ValueError("q must be of type either Sample or torch.Tensor")
        self.p = p

    @property
    def point(
            self
    ) -> torch.Tensor:

        """
        Getter for the `point` property.

        Returns
        ----------
        torch.Tensor
            The value of the position sample.
        """
        
        return self.q.point

    @point.setter
    def point(
            self,
            value : torch.Tensor
    ):

        """
        Setter for the `point` property.

        Parameters
        ----------
        value : torch.Tensor
            The position value of the Hamiltonian system.
        """
        
        self.q.point = value

    @property
    def log_prob(
            self
    ) -> torch.Tensor:

        """
        Getter for the `log_prob` property.

        Returns
        ----------
        torch.Tensor
            The log probability of a distribution of interested evaluated at the value of `self.point`.
        """
        
        return self.q.log_prob

    @log_prob.setter
    def log_prob(
            self,
            value : torch.Tensor
    ):

        """
        Setter for the `log_prob` property.

        Parameters
        ----------
        value : torch.Tensor
            The log probability of a distribution of interested evaluated at `self.point`.
        """
        
        self.q.log_prob = value
